import argparse

__author__ = 'Kapustlo'

__version__ = '1.2.1'

parser = argparse.ArgumentParser(description='Build Django project')

parser.add_argument('name', metavar='name', help='Django project name')

parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')

parser.add_argument('--dir', dest='dir', type=str, default='./', required=False, help='Django project directory')

parser.add_argument('--debug', dest='debug', action='store_true', required=False, help='Should create env with DEBUG=True')

parser.add_argument('--hosts', dest='hosts', required=False, default='', help='List of comma separated ALLOWED_HOSTS values')

parser.add_argument('--python', dest='python', required=False, default='python3', help='Python command')

parser.add_argument('--migrate', dest='migrate', required=False, action='store_true', help='Apply default migrations after creation')

parser.add_argument('--git', dest='git', required=False, action='store_true', help='Initialize git repo')

parser.add_argument('--commit', dest='commit', required=False, action='store_true', help='Make initial git commit')

parser.add_argument('--commit-message', dest='commit_message', required=False, default='Initial commit', help='Initial commit name')

parser.add_argument('--packages', dest='packages', nargs='+', help='Additional pip packages', required=False, default=[])



import sys
import asyncio
import secrets
from subprocess import SubprocessError
from asyncio.subprocess import PIPE, Process
from typing import Tuple, AnyStr, List, Iterable

import aiofiles

SYS_ENCODING = sys.getdefaultencoding()


def generate_env(debug: bool=False, hosts: str='*') -> str:
    items = [
        f'SECRET_KEY={secrets.token_hex(128)}',
        f'ALLOWED_HOSTS={hosts}'
    ]

    if debug:
        items.append('DEBUG=True')

    return '\n\n'.join(items)


def edit_settings(lines: Iterable[str]) -> List[str]:
    result = []

    for line in lines:
        if line.startswith('BASE_DIR'):
            result.append('\nfrom environ import Env\n')

            result.append(f'\n{line}\n')

            result.append('\nenv = Env(DEBUG=(bool, False))\n')
            
            result.append('\nEnv.read_env(BASE_DIR / \'.env\')\n')

        elif line.startswith('SECRET_KEY'):
            result.append(f'SECRET_KEY = env(\'SECRET_KEY\')\n')
        elif line.startswith('ALLOWED_HOSTS'):
            result.append('ALLOWED_HOSTS = env.tuple(\'ALLOWED_HOSTS\')')
        elif line.startswith('DEBUG'):
            result.append('DEBUG = env(\'DEBUG\')\n')
        else:
            result.append(line)

    return result


async def write_file(path: str, content: AnyStr, mode: str='w'):
    async with aiofiles.open(path, mode) as file:
        await file.write(content)


async def run_cmd(args: Tuple[str], **kwargs) -> Process:
    cmd = ' '.join(args)

    kwargs = {
        'stdout': PIPE,
        'stderr': PIPE,
        **kwargs
    }

    result = await asyncio.create_subprocess_shell(cmd, **kwargs)

    out, err = [item.decode(SYS_ENCODING) if isinstance(item, bytes) else item for item in await result.communicate()]

    code = result.returncode

    if code != 0:
        raise SubprocessError(f'{code}: {err}')

    if out:
        print(out)

    return result


async def install_packages(pip_path: str, packages: Iterable[str]):
    tasks = [
        run_cmd((
            pip_path,
            'install',
            '--require-virtualenv',
            '-v',
            '--disable-pip-version-check',
            package
        )) for package in packages
    ]

    await asyncio.gather(*tasks)



import os
import asyncio
from shlex import quote

import aiofiles

from . import  const, utils, parser, __version__


async def exec():
    args = parser.parse_args()

    project_dir = args.dir

    print('Checking project directory...')

    if not os.path.exists(project_dir):
        print('Project directory not found, creating project directory')

        os.mkdir(project_dir)
    else:
        print('Project directory found')

    name = args.name 

    print('Creating Django project...')

    await utils.run_cmd(('django-admin', 'startproject', name, project_dir))

    print('Django project created')

    venv_dir = os.path.join(project_dir, 'venv/')

    print('Creating files...')

    tasks = [
        utils.write_file(
            os.path.join(project_dir, '.gitignore'),
            const.GITIGNORE
        ),
        utils.write_file(
            os.path.join(project_dir, '.env'),
            utils.generate_env(args.debug, args.hosts)
        ),
        utils.write_file(
            os.path.join(project_dir, 'env.example'),
            const.ENV_EXAMPLE
        ),
        utils.run_cmd([args.python, '-m', 'venv', quote(venv_dir)])
    ]

    if args.git:
        tasks.append(utils.run_cmd(['git', 'init', project_dir]))  

    await asyncio.gather(*tasks)

    print('Created files')

    bin_path = os.path.join(venv_dir, 'bin')

    pip_path = os.path.join(bin_path, 'pip') 

    print('Installing required packages...')

    packages = map(quote, frozenset((*const.REQUIRED_PACKAGES, *args.packages)))

    await utils.install_packages(pip_path, packages)

    print('Installed required packages')

    print('Writing requirements...')

    requirements_path = os.path.join(project_dir, 'requirements.txt')

    with open(requirements_path, 'w') as file:
        await utils.run_cmd((pip_path, 'freeze'), stdout=file)

    print('Added requirements')

    print('Editing settings file...')

    settings_dir = os.path.join(project_dir, args.name, 'settings.py')

    async with aiofiles.open(settings_dir, 'r+') as file:
        lines = await file.readlines()

        result = utils.edit_settings(lines)

        await file.seek(0)

        await file.writelines(result)

    print('Edited settings file')
    
    tasks = []

    if args.git and args.commit:
        tasks.append(utils.run_cmd((
            'cd',
            quote(project_dir),
            '&&',
            'git',
            'add',
            '-v',
            '.',
            '&&'
            'git',
            'commit',
            '-v',
            '-m',
            f'"{args.commit_message}"'
        )))

    if args.migrate:
        python_path = os.path.join(bin_path, 'python')

        manage_path = os.path.join(project_dir, 'manage.py')

        tasks.append(utils.run_cmd((
            python_path,
            quote(manage_path),
            'migrate',
            '--no-input'
        )))

    await asyncio.gather(*tasks)

    
def main():
    asyncio.run(exec())

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        exit()
